"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import TYPE_CHECKING

from encommon.types import DictStrAny

from .child import RobieChild
from ..addons import RobieQueue

if TYPE_CHECKING:
    from ..models import RobieCommand
    from ..models import RobieMessage



class RobieClient(RobieChild):
    """
    Establish and maintain connection with the chat service.
    """


    @property
    def family(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def kind(
        self,
    ) -> Literal['client']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'client'


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        dumped = super().dumped

        family = self.family

        return dumped | {
            'family': family}


    def get_message(
        self,
    ) -> 'RobieMessage':
        """
        Return the new item containing information for operation.

        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def put_message(
        self,
        mqueue: RobieQueue['RobieMessage'],
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param mqueue: Queue instance where the item is received.
        """

        raise NotImplementedError


    def get_command(
        self,
    ) -> 'RobieCommand':
        """
        Return the new item containing information for operation.

        :returns: New item containing information for operation.
        """

        raise NotImplementedError


    def put_command(
        self,
        cqueue: RobieQueue['RobieCommand'],
    ) -> None:
        """
        Insert the new item containing information for operation.

        :param cqueue: Queue instance where the item is received.
        """

        raise NotImplementedError


    def compose(
        self,
        target: str,
        content: str,
    ) -> 'RobieCommand':
        """
        Compose the message and transmit using the chat client.

        :param target: Where the composed message will be sent.
        :param content: Content that will be source of message.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING

from encommon.types import getate
from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs
from encommon.types import setate

from enconnect.discord.test import EVENTS as DSCEVENTS
from enconnect.fixtures import DSCClientSocket
from enconnect.fixtures import IRCClientSocket
from enconnect.fixtures import MTMClientSocket
from enconnect.irc.test import EVENTS as IRCEVENTS
from enconnect.mattermost.test import EVENTS as MTMEVENTS

from ..plugin import StatusPlugin

if TYPE_CHECKING:
    from ....robie import Robie
    from ....robie import RobieService



DSCEVENTS = deepcopy(DSCEVENTS)
IRCEVENTS = deepcopy(IRCEVENTS)
MTMEVENTS = deepcopy(MTMEVENTS)



setate(
    DSCEVENTS[0],
    'd/content',
    '!status')

setate(
    DSCEVENTS[1],
    'd/content',
    '!status')



IRCEVENTS[1] = (
    IRCEVENTS[1]
    .replace(
        'Hello ircbot',
        '!status'))

IRCEVENTS[3] = (
    IRCEVENTS[3]
    .replace(
        'Hello world',
        '!status'))



setate(
    MTMEVENTS[0],
    'data/post',
    (getate(
        MTMEVENTS[0],
        'data/post')
     .replace(
         'Hello mtmbot',
         '!status')))

setate(
    MTMEVENTS[1],
    'data/post',
    (getate(
        MTMEVENTS[1],
        'data/post')
     .replace(
         'Hello world',
         '!status')))



def test_StatusPlugin(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    plugins = childs.plugins


    plugin = plugins['status']

    assert isinstance(
        plugin, StatusPlugin)


    attrs = lattrs(plugin)

    assert attrs == [
        '_RobieChild__robie',
        '_RobieChild__name',
        '_RobieChild__params']


    assert inrepr(
        'plugin.StatusPlugin',
        plugin)

    assert hash(plugin) > 0

    assert instr(
        'plugin.StatusPlugin',
        plugin)


    plugin.validate()

    assert plugin.robie

    assert plugin.enable

    assert plugin.name == 'status'

    assert plugin.kind == 'plugin'

    assert plugin.params

    assert plugin.dumped



def test_StatusPlugin_cover(
    service: 'RobieService',
    client_dscsock: DSCClientSocket,
    client_ircsock: IRCClientSocket,
    client_mtmsock: MTMClientSocket,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    :param client_dscsock: Object to mock client connection.
    :param client_ircsock: Object to mock client connection.
    :param client_mtmsock: Object to mock client connection.
    """

    client_dscsock(DSCEVENTS)
    client_ircsock(IRCEVENTS)
    client_mtmsock(MTMEVENTS)

    service.start()


    thread = Thread(
        target=service.operate)

    thread.start()


    block_sleep(10)

    service.soft()

    while service.running:
        block_sleep(1)

    service.stop()

    thread.join()

    assert not service.congest



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from encommon.types import DictStrAny
from encommon.types import NCNone

from .client import RobieClient
from .plugin import RobiePlugin

if TYPE_CHECKING:
    from ..robie import Robie



RobieClients = dict[str, RobieClient]
RobiePlugins = dict[str, RobiePlugin]



class RobieChilds:
    """
    Contain the object instances for related Robie children.

    :param robie: Primary class instance for Chatting Robie.
    """

    __robie: 'Robie'

    __clients: RobieClients
    __plugins: RobiePlugins


    def __init__(
        self,
        robie: 'Robie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        robie.logger.log_d(
            base=self,
            status='initial')

        self.__robie = robie

        self.__clients = {}
        self.__plugins = {}

        robie.logger.log_i(
            base=self,
            status='created')


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """


        clients = (
            self.__clients
            .values())

        for client in clients:
            client.validate()


        plugins = (
            self.__plugins
            .values())

        for plugin in plugins:
            plugin.validate()


    @property
    def clients(
        self,
    ) -> RobieClients:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        clients = self.__clients

        return dict(clients)


    @property
    def plugins(
        self,
    ) -> RobiePlugins:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        plugins = self.__plugins

        return dict(plugins)


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        clients = self.clients
        plugins = self.plugins

        dumped: DictStrAny = {

            'clients': {
                k: v.dumped for k, v
                in clients.items()},

            'plugins': {
                k: v.dumped for k, v
                in plugins.items()}}

        return deepcopy(dumped)


    def register(  # noqa: CFQ004
        self,
        name: str,
        *,
        client: Optional[Type[RobieClient]] = None,
        plugin: Optional[Type[RobiePlugin]] = None,
    ) -> None:
        """
        Register the plugin with the internal operation routine.

        :param name: Name of the object within the Robie config.
        :param client: Class definition for the instantiation.
        :param plugin: Class definition for the instantiation.
        """

        robie = self.__robie
        params = robie.params

        _clients = params.clients
        _plugins = params.plugins


        def _put_client() -> None:

            target = self.__clients

            if _clients is None:
                return NCNone

            if name not in _clients:
                return NCNone

            params = _clients[name]

            assert callable(client)

            if not params.enable:
                return NCNone

            object = client(
                robie, name, params)

            target[name] = object


        if client is not None:
            _put_client()


        def _put_plugin() -> None:

            target = self.__plugins

            if _plugins is None:
                return NCNone

            if name not in _plugins:
                return NCNone

            params = _plugins[name]

            assert callable(plugin)

            if not params.enable:
                return NCNone

            object = plugin(
                robie, name, params)

            target[name] = object


        if plugin is not None:
            _put_plugin()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Literal
from typing import TYPE_CHECKING

from ...robie.models import RobieCommand

if TYPE_CHECKING:
    from ...robie.childs import RobieClient



@dataclass
class IRCCommand(RobieCommand):
    """
    Contain information for sharing using the Python queue.
    """

    event: str


    def __init__(
        self,
        client: 'RobieClient',
        event: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        # Translate to client.command

        self.event = event

        super().__init__(client=client)


    @property
    def family(
        self,
    ) -> Literal['irc']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'irc'



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Event
from typing import TYPE_CHECKING

from encommon.types import clsname

from ..addons import RobieQueue
from ..threads import RobieThread  # noqa: F401

if TYPE_CHECKING:
    from ..models import RobieCommand
    from ..models import RobieMessage
    from ..robie import Robie



RobieThreads = dict[str, 'RobieThread']

_CONGEST = list[tuple[str, int]]



class RobieMember:
    """
    Common methods and routines for Chatting Robie members.

    :param robie: Primary class instance for Chatting Robie.
    """

    __robie: 'Robie'

    __threads: RobieThreads

    __mqueue: RobieQueue['RobieMessage']
    __cqueue: RobieQueue['RobieCommand']
    __vacate: Event
    __cancel: Event


    def __init__(
        self,
        robie: 'Robie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        robie.logger.log_d(
            base=self,
            status='initial')

        self.__robie = robie

        self.__threads = {}

        self.__build_objects()

        robie.logger.log_i(
            base=self,
            status='created')


    def __build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        robie = self.__robie

        self.__mqueue = (
            RobieQueue(robie))

        self.__cqueue = (
            RobieQueue(robie))

        self.__vacate = Event()
        self.__cancel = Event()

        self.build_threads()


    def build_threads(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.

        .. note::
           Deviates from enhomie in build happens downstream,
           because all of the enhomie members are origin based
           where enrobie children do not have consistent base.
        """

        raise NotImplementedError


    @property
    def robie(
        self,
    ) -> 'Robie':
        """
        Return the Robie instance to which this instance belongs.

        :returns: Robie instance to which this instance belongs.
        """

        return self.__robie


    @property
    def threads(
        self,
    ) -> RobieThreads:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        threads = self.__threads

        return dict(threads)


    @threads.setter
    def threads(
        self,
        value: RobieThreads,
    ) -> None:
        """
        Update the value for the attribute from class instance.
        """

        self.__threads = value


    @property
    def mqueue(
        self,
    ) -> RobieQueue['RobieMessage']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__mqueue


    @property
    def cqueue(
        self,
    ) -> RobieQueue['RobieCommand']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__cqueue


    @property
    def vacate(
        self,
    ) -> Event:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__vacate


    @property
    def cancel(
        self,
    ) -> Event:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__cancel


    @property
    def running(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined running.

        :returns: List of threads which are determined running.
        """

        running: list[str] = []

        threads = (
            self.__threads
            .values())

        for thread in threads:

            if not thread.is_alive():
                continue

            name = thread.name

            running.append(name)

        return sorted(running)


    @property
    def zombies(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined zombies.

        :returns: List of threads which are determined zombies.
        """

        zombies: list[str] = []

        threads = (
            self.__threads
            .values())

        for thread in threads:

            if thread.is_alive():
                continue

            name = thread.name

            zombies.append(name)

        return sorted(zombies)


    @property
    def congest(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        congest: _CONGEST = []

        name = clsname(self)

        mqueue = self.mqueue
        cqueue = self.cqueue


        if mqueue.qsize > 5:

            append = (
                f'{name}/mqueue',
                mqueue.qsize)

            congest.append(append)

        if cqueue.qsize > 5:

            append = (
                f'{name}/cqueue',
                cqueue.qsize)

            congest.append(append)


        threads = (
            self.__threads
            .values())

        for thread in threads:

            congest.extend(
                thread.congest)


        return sorted(congest)


    @property
    def enqueue(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        enqueue: _CONGEST = []

        name = clsname(self)

        mqueue = self.mqueue
        cqueue = self.cqueue


        if mqueue.qsize > 0:

            append = (
                f'{name}/mqueue',
                mqueue.qsize)

            enqueue.append(append)

        if cqueue.qsize > 0:

            append = (
                f'{name}/cqueue',
                cqueue.qsize)

            enqueue.append(append)


        threads = (
            self.__threads
            .values())

        for thread in threads:

            enqueue.extend(
                thread.enqueue)


        return sorted(enqueue)


    def start(
        self,
    ) -> None:
        """
        Start the various threads within the Robie class object.
        """

        threads = (
            self.__threads
            .values())

        for thread in threads:
            thread.start()


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Robie service members.
        """

        raise NotImplementedError


    def soft(
        self,
    ) -> None:
        """
        Stop the various threads within the Robie class object.
        """

        self.__vacate.set()


    def stop(
        self,
    ) -> None:
        """
        Stop the various threads within the Robie class object.
        """

        self.__vacate.set()
        self.__cancel.set()

        threads = (
            self.__threads
            .values())

        for thread in threads:
            thread.stop()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Literal
from typing import TYPE_CHECKING

from ..addons import RobieQueueItem

if TYPE_CHECKING:
    from ..childs import RobieClient
    from ..models import RobieCommand
    from ..robie import Robie



RobieMessageKinds = Literal[
    'event',
    'chanmsg',
    'privmsg']



@dataclass
class RobieMessage(RobieQueueItem):
    """
    Contain information for sharing using the Python queue.
    """

    client: str


    def __init__(
        self,
        client: 'RobieClient',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.client = client.name

        super().__init__()


    @property
    def family(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def kind(
        self,
    ) -> 'RobieMessageKinds':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def isme(
        self,
    ) -> bool:
        """
        Return the boolean indicating message origin from client.

        :returns: Boolean indicating message origin from client.
        """

        raise NotImplementedError


    def reply(
        self,
        robie: 'Robie',
        content: str,
    ) -> 'RobieCommand':
        """
        Compose the message and transmit using the chat client.

        :param robie: Primary class instance for Chatting Robie.
        :param content: Content that will be source of message.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import DictStrAny

if TYPE_CHECKING:
    from ..common import RobieKinds
    from ..params import RobieChildParams
    from ..robie import Robie
    from ..threads import RobieThread



class RobieChild:
    """
    Normalize the desired parameters with multiple products.

    :param robie: Primary class instance for Chatting Robie.
    :param name: Name of the object within the Robie config.
    :param params: Parameters used to instantiate the class.
    """

    __robie: 'Robie'

    __name: str
    __params: 'RobieChildParams'


    def __init__(
        self,
        robie: 'Robie',
        name: str,
        params: 'RobieChildParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        robie.logger.log_d(
            base=self,
            name=name,
            status='initial')

        self.__robie = robie
        self.__name = name
        self.__params = params

        self.__post__()

        robie.logger.log_d(
            base=self,
            name=name,
            status='created')


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        raise NotImplementedError


    def __lt__(
        self,
        other: 'RobieChild',
    ) -> bool:
        """
        Built-in method for comparing this instance with another.

        .. note::
           Useful with sorting to influence consistent output.

        :param other: Other value being compared with instance.
        :returns: Boolean indicating outcome from the operation.
        """

        name = self.name
        _name = other.name

        return name < _name


    @property
    def robie(
        self,
    ) -> 'Robie':
        """
        Return the Robie instance to which this instance belongs.

        :returns: Robie instance to which this instance belongs.
        """

        return self.__robie


    @property
    def enable(
        self,
    ) -> bool:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.params.enable


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    @property
    def kind(
        self,
    ) -> 'RobieKinds':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        raise NotImplementedError


    @property
    def params(
        self,
    ) -> 'RobieChildParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def dumped(
        self,
    ) -> DictStrAny:
        """
        Return the facts about the attributes from the instance.

        :returns: Facts about the attributes from the instance.
        """

        params = self.__params
        dumped = params.endumped

        return {
            'name': self.name,
            'kind': self.kind,
            'params': dumped}


    def operate(
        self,
        thread: 'RobieThread',
    ) -> None:
        """
        Perform the operation related to Robie service threads.

        .. note::
           Deviates from enhomie in children have operations,
           and are more isolated from internal core routines.

        :param thread: Child class instance for Chatting Robie.
        """

        raise NotImplementedError



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...service import RobieService



def test_RobiePlugins(
    service: 'RobieService',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param service: Ancilary Chatting Robie class instance.
    """

    member = service.plugins


    attrs = lattrs(member)

    assert attrs == [
        '_RobieMember__robie',
        '_RobieMember__threads',
        '_RobieMember__mqueue',
        '_RobieMember__cqueue',
        '_RobieMember__vacate',
        '_RobieMember__cancel']


    assert inrepr(
        'plugins.RobiePlugins',
        member)

    assert hash(member) > 0

    assert instr(
        'plugins.RobiePlugins',
        member)


    assert member.robie

    assert len(member.threads) == 1

    assert member.mqueue

    assert member.cqueue

    assert member.vacate

    assert member.cancel

    assert len(member.running) == 0


    member.operate()



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from time import sleep as block_sleep
from typing import TYPE_CHECKING
from typing import Union

from encommon.types import clsname

from ..addons import RobieQueue

if TYPE_CHECKING:
    from ..childs import RobieChild
    from ..members import RobieMember
    from ..models import RobieCommand
    from ..models import RobieMessage
    from ..robie import Robie



_CONGEST = list[tuple[str, int]]



RobieThreadItems = Union[
    'RobieMessage',
    'RobieCommand']



class RobieThread(Thread):
    """
    Common methods and routines for Chatting Robie threads.

    :param member: Child class instance for Chatting Robie.
    :param child: Child class instance for Chatting Robie.
    """

    __member: 'RobieMember'
    __child: 'RobieChild'

    __mqueue: RobieQueue['RobieMessage']
    __cqueue: RobieQueue['RobieCommand']


    def __init__(
        self,
        member: 'RobieMember',
        child: 'RobieChild',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        robie = member.robie

        super().__init__()

        self.name = (
            f'{clsname(self)}'
            f'/{child.name}')

        robie.logger.log_d(
            base=self,
            name=child,
            status='initial')

        self.__member = member
        self.__child = child

        self.__build_objects()

        robie.logger.log_i(
            base=self,
            name=child,
            status='created')


    def __build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        member = self.__member
        robie = member.robie

        self.__mqueue = (
            RobieQueue(robie))

        self.__cqueue = (
            RobieQueue(robie))


    @property
    def robie(
        self,
    ) -> 'Robie':
        """
        Return the Robie instance to which this instance belongs.

        :returns: Robie instance to which this instance belongs.
        """

        return self.member.robie


    @property
    def member(
        self,
    ) -> 'RobieMember':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__member


    @property
    def child(
        self,
    ) -> 'RobieChild':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__child


    @property
    def mqueue(
        self,
    ) -> RobieQueue['RobieMessage']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__mqueue


    @property
    def cqueue(
        self,
    ) -> RobieQueue['RobieCommand']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__cqueue


    def expired(
        self,
        item: RobieThreadItems,
    ) -> bool:
        """
        Return the boolean indicating whether the item expired.

        :param item: Item containing information for operation.
        :returns: Boolean indicating whether the item expired.
        """

        member = self.member
        robie = member.robie

        since = item.time.since
        client = item.client

        if since < 60:
            return False

        robie.logger.log_w(
            base=self,
            item=item,
            client=client,
            status='expired',
            age=int(since))

        return True


    @property
    def congest(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        congest: _CONGEST = []

        name = self.name

        mqueue = self.mqueue
        cqueue = self.cqueue


        if mqueue.qsize > 5:

            append = (
                f'{name}/mqueue',
                mqueue.qsize)

            congest.append(append)

        if cqueue.qsize > 5:

            append = (
                f'{name}/cqueue',
                cqueue.qsize)

            congest.append(append)


        return sorted(congest)


    @property
    def enqueue(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        enqueue: _CONGEST = []

        name = self.name

        mqueue = self.mqueue
        cqueue = self.cqueue


        if mqueue.qsize > 0:

            append = (
                f'{name}/mqueue',
                mqueue.qsize)

            enqueue.append(append)

        if cqueue.qsize > 0:

            append = (
                f'{name}/cqueue',
                cqueue.qsize)

            enqueue.append(append)


        return sorted(enqueue)


    def run(
        self,
    ) -> None:
        """
        Perform whatever operation is associated with the class.
        """

        member = self.__member
        child = self.__child
        robie = member.robie
        vacate = member.vacate
        cancel = member.cancel

        robie.logger.log_i(
            base=self,
            name=child,
            status='started')


        def _continue() -> bool:

            if cancel.is_set():
                return False

            enqueue = bool(
                self.enqueue)

            if vacate.is_set():
                return enqueue

            return True


        while _continue():

            self.operate()

            block_sleep(0.1)


        robie.logger.log_i(
            base=self,
            name=child,
            status='stopped')


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service threads.
        """

        member = self.__member
        robie = member.robie
        child = self.__child

        try:
            child.operate(self)

        except Exception as reason:

            robie.logger.log_e(
                base=self,
                name=child,
                status='exception',
                exc_info=reason)

            block_sleep(1)


    def stop(
        self,
    ) -> None:
        """
        Wait for the thread object to complete routine and exit.
        """

        member = self.__member
        child = self.__child
        robie = member.robie

        robie.logger.log_d(
            base=self,
            name=child,
            status='waiting')

        self.join()

        robie.logger.log_d(
            base=self,
            name=child,
            status='awaited')



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Annotated
from typing import Any
from typing import Callable
from typing import Optional

from enconnect.irc import ClientParams

from pydantic import Field

from ...robie.params import RobieClientParams



class IRCClientParams(RobieClientParams, extra='forbid'):
    """
    Process and validate the Robie configuration parameters.
    """

    client: Annotated[
        ClientParams,
        Field(...,
              description='Parameters for the base client')]


    def __init__(
        # NOCVR
        self,
        /,
        _parse: Optional[Callable[..., Any]] = None,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """


        if _parse is not None:

            parsable = ['client']

            for key in parsable:

                value = data.get(key)

                if value is None:
                    continue

                value = _parse(value)

                data[key] = value


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..queue import RobieQueue
from ..queue import RobieQueueItem

if TYPE_CHECKING:
    from ...robie import Robie



def test_RobieQueueItem() -> None:
    """
    Perform various tests associated with relevant routines.
    """

    item = RobieQueueItem()

    assert item.time.since < 1



def test_RobieQueue(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    _queue = RobieQueue[
        RobieQueueItem]

    item = RobieQueueItem()


    queue: _queue = (
        RobieQueue(robie))


    attrs = lattrs(queue)

    assert attrs == [
        '_RobieQueue__robie',
        '_RobieQueue__queue']


    assert inrepr(
        'queue.RobieQueue',
        queue)

    assert hash(queue) > 0

    assert instr(
        'queue.RobieQueue',
        queue)


    assert queue.empty

    assert queue.qsize == 0


    queue.put(item)

    assert not queue.empty

    assert queue.qsize == 1

    _item = queue.get()

    assert queue.empty

    assert queue.qsize == 0

    assert item == _item



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from time import sleep as block_sleep
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from enconnect.discord import Client
from enconnect.discord import ClientEvent
from enconnect.utils.http import _METHODS
from enconnect.utils.http import _PAYLOAD

from .command import DSCCommand
from .message import DSCMessage
from .params import DSCClientParams
from ...robie.addons import RobieQueue
from ...robie.childs import RobieClient

if TYPE_CHECKING:
    from ...robie.models import RobieCommand
    from ...robie.models import RobieMessage
    from ...robie.threads import RobieThread



class DSCClient(RobieClient):
    """
    Establish and maintain connection with the chat service.

    :param robie: Primary class instance for Chatting Robie.
    """

    __client: Client


    def __post__(
        self,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        params = self.params

        assert isinstance(
            params, DSCClientParams)

        client = Client(
            params.client,
            self.__debugger)

        self.__client = client


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """


    @property
    def family(
        self,
    ) -> Literal['discord']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'discord'


    @property
    def client(
        self,
    ) -> Client:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__client


    def operate(  # noqa: CFQ001
        self,
        thread: 'RobieThread',
    ) -> None:
        """
        Perform the operation related to Homie service threads.

        :param thread: Child class instance for Chatting Robie.
        """

        robie = thread.robie
        member = thread.member
        target = member.mqueue
        cqueue = thread.cqueue
        vacate = member.vacate
        params = self.params

        assert isinstance(
            params, DSCClientParams)


        def _put_mqueue() -> None:

            event = source.get()

            self.put_message(
                target, event)


        def _get_cqueue() -> None:

            citem = cqueue.get()

            assert isinstance(
                citem, DSCCommand)

            client.request(
                method=citem.method,
                path=citem.path,
                params=citem.params,
                json=citem.json,
                timeout=10)


        def _continue() -> bool:

            return (
                not vacate.is_set()
                and daerht.is_alive())


        client = self.__client
        source = client.mqueue


        def _operate() -> None:

            try:
                client.operate()

            except ConnectionError:
                return None

            except Exception as reason:

                robie.logger.log_e(
                    base=self,
                    name=self,
                    status='exception',
                    exc_info=reason)

                return None


        def _routine() -> None:

            while _continue():

                robie.logger.log_i(
                    base=self,
                    name=self,
                    status='connect')

                _operate()

                robie.logger.log_i(
                    base=self,
                    name=self,
                    status='severed')

                block_sleep(1)


        daerht = Thread(
            target=_routine)

        daerht.start()


        while _continue():

            if not cqueue.empty:
                _get_cqueue()

            while not source.empty():
                _put_mqueue()

            block_sleep(0.05)


        client.stop()

        while daerht.is_alive():
            daerht.join(1)

        while not source.empty():
            _put_mqueue()  # NOCVR


    def get_message(
        self,
        event: Optional[ClientEvent] = None,
    ) -> DSCMessage:
        """
        Return the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param event: Raw event received from the network peer.
        :returns: New item containing information for operation.
        """

        assert event is not None

        return DSCMessage(self, event)


    def put_message(
        self,
        mqueue: RobieQueue['RobieMessage'],
        event: Optional[ClientEvent] = None,
    ) -> None:
        """
        Insert the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param mqueue: Queue instance where the item is received.
        :param event: Raw event received from the network peer.
        """

        mitem = self.get_message(event)

        mqueue.put(mitem)


    def get_command(
        self,
        method: Optional[_METHODS] = None,
        path: Optional[str] = None,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
    ) -> DSCCommand:
        """
        Return the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :returns: New item containing information for operation.
        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        """

        assert method is not None
        assert path is not None

        return DSCCommand(
            self,
            method, path,
            params, json)


    def put_command(
        self,
        cqueue: RobieQueue['RobieCommand'],
        method: Optional[_METHODS] = None,
        path: Optional[str] = None,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
    ) -> None:
        """
        Insert the new item containing information for operation.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param cqueue: Queue instance where the item is received.
        :param method: Method for operation with the API server.
        :param path: Path for the location to upstream endpoint.
        :param params: Optional parameters included in request.
        :param json: Optional JSON payload included in request.
        """

        citem = (
            self.get_command(
                method, path,
                params, json))

        cqueue.put(citem)


    def compose(
        self,
        target: str,
        content: str,
    ) -> 'RobieCommand':
        """
        Compose the message and transmit using the chat client.

        .. note::
           Though parameters are marked optional, that is not
           necessarily the case; for typing purposes one cannot
           introduce required parameters outside of the parent.

        :param target: Where the composed message will be sent.
        :param content: Content that will be source of message.
        """

        assert target is not None

        path = (
            f'channels/{target}'
            '/messages')

        payload = {
            'content': content}

        return (
            self.get_command(
                'post', path,
                json=payload))


    def __debugger(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        :param kwargs: Keyword arguments for populating message.
        """

        robie = self.robie

        assert not any([
            kwargs.get('level'),
            kwargs.get('base'),
            kwargs.get('name')])

        robie.logger.log_d(
            base=self,
            name=self,
            **kwargs)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...robie import Robie



def test_RobiePlugin_cover(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    # Remove if not needed



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from ..param import InvalidParam

if TYPE_CHECKING:
    from ...robie import Robie



def test_InvalidParam(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']


    raises = InvalidParam(
        error='invalid',
        about='about',
        child=client,
        param='param',
        value='value')


    attrs = lattrs(raises)

    assert attrs == [
        'error',
        'about',
        'child',
        'param',
        'value']


    assert inrepr(
        'InvalidParam',
        raises)

    assert hash(raises) > 0

    assert instr(
        'Error (invalid)',
        raises)


    assert str(raises) == (
        'Error (invalid) '
        'param (param) '
        'value (value) child '
        '(IRCClient/ircbot)'
        ' (about)')



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .child import RobieChildParams
from .client import RobieClientParams
from .plugin import RobiePluginParams
from .robie import RobieParams
from .robie import RobiePrinterParams
from .service import RobieServiceParams



__all__ = [
    'RobieParams',
    'RobiePrinterParams',
    'RobieChildParams',
    'RobiePluginParams',
    'RobieClientParams',
    'RobieServiceParams']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .client import RobieClientThread
from .plugin import RobiePluginThread
from .thread import RobieThread
from .thread import RobieThreadItems



__all__ = [
    'RobieThread',
    'RobieThreadItems',
    'RobieClientThread',
    'RobiePluginThread']



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from pytest import raises

from ..command import RobieCommand

if TYPE_CHECKING:
    from ...robie import Robie



def test_RobieCommand(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients
    client = clients['ircbot']

    model = RobieCommand


    item = model(client)


    attrs = lattrs(item)

    assert attrs == [
        'client',
        'time']


    assert inrepr(
        'RobieCommand',
        item)

    with raises(TypeError):
        assert hash(item) > 0

    assert instr(
        'RobieCommand',
        item)


    assert item.time.since > 0

    assert item.client == client.name



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from encommon.types import DictStrAny
from encommon.utils import save_text

from enconnect.fixtures import client_dscsock
from enconnect.fixtures import client_ircsock
from enconnect.fixtures import client_mtmsock

from pytest import fixture

from . import EXAMPLES
from . import PROJECT
from .clients import DSCClient
from .clients import DSCClientParams
from .clients import IRCClient
from .clients import IRCClientParams
from .clients import MTMClient
from .clients import MTMClientParams
from .plugins import StatusPlugin
from .plugins import StatusPluginParams
from .robie import Robie
from .robie import RobieConfig
from .robie import RobieService



__all__ = [
    'client_dscsock',
    'client_ircsock',
    'client_mtmsock']



def config_factory(
    tmp_path: Path,
) -> RobieConfig:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    content = (
        f"""

        enconfig:
          paths:
            - {EXAMPLES}
            - {tmp_path}/robie

        enlogger:
          stdo_level: info

        database: >-
            sqlite:///{tmp_path}/db

        """)  # noqa: LIT003

    config_path = (
        tmp_path / 'config.yml')

    save_text(
        config_path, content)

    sargs = {
        'config': config_path,
        'console': True,
        'debug': True,
        'pmessage': True,
        'pcommand': True}

    return RobieConfig(sargs)



@fixture
def config(
    tmp_path: Path,
) -> RobieConfig:
    """
    Construct the instance for use in the downstream tests.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Newly constructed instance of related class.
    """

    config = config_factory(tmp_path)

    config.register(
        name='dscbot',
        client=DSCClientParams)

    config.register(
        name='ircbot',
        client=IRCClientParams)

    config.register(
        name='mtmbot',
        client=MTMClientParams)

    config.register(
        name='status',
        plugin=StatusPluginParams)

    return config



@fixture
def replaces(
    tmp_path: Path,
) -> DictStrAny:
    """
    Return the complete mapping of what replaced in sample.

    :param tmp_path: pytest object for temporal filesystem.
    :returns: Complete mapping of what replaced in sample.
    """

    return {
        'PROJECT': PROJECT,
        'TMPPATH': tmp_path}



def robie_factory(
    config: RobieConfig,
) -> Robie:
    """
    Construct the instance for use in the downstream tests.

    :param config: Primary class instance for configuration.
    :returns: Newly constructed instance of related class.
    """

    robie = Robie(config)

    robie.register(
        name='dscbot',
        client=DSCClient)

    robie.register(
        name='ircbot',
        client=IRCClient)

    robie.register(
        name='mtmbot',
        client=MTMClient)

    robie.register(
        name='status',
        plugin=StatusPlugin)

    return robie



@fixture
def robie(
    config: RobieConfig,
) -> Robie:
    """
    Construct the instance for use in the downstream tests.

    :param config: Primary class instance for configuration.
    :returns: Newly constructed instance of related class.
    """

    return robie_factory(config)



def service_factory(
    robie: Robie,
) -> RobieService:
    """
    Construct the instance for use in the downstream tests.

    :param robie: Primary class instance for Chatting Robie.
    :returns: Newly constructed instance of related class.
    """

    return RobieService(robie)



@fixture
def service(
    robie: Robie,
) -> RobieService:
    """
    Construct the instance for use in the downstream tests.

    :param robie: Primary class instance for Chatting Robie.
    :returns: Newly constructed instance of related class.
    """

    return service_factory(robie)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Thread
from typing import TYPE_CHECKING

from _pytest.logging import LogCaptureFixture

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

if TYPE_CHECKING:
    from ...robie import Robie



def test_RobieLogger(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    logger = robie.logger


    attrs = lattrs(logger)

    assert attrs == [
        '_RobieLogger__robie']


    assert inrepr(
        'logger.RobieLogger',
        logger)

    assert hash(logger) > 0

    assert instr(
        'logger.RobieLogger',
        logger)



def test_RobieLogger_message(
    robie: 'Robie',
    caplog: LogCaptureFixture,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    :param caplog: pytest object for capturing log message.
    """

    logger = robie.logger


    logger.start()

    logger.log_d(about='pytest')
    logger.log_c(about='pytest')
    logger.log_e(about='pytest')
    logger.log_i(about='pytest')
    logger.log_w(about='pytest')

    logger.log(
        level='debug',
        about='pytest')

    logger.stop()

    output = caplog.record_tuples

    assert len(output) == 6

    logger.log_d(about='pytest')
    logger.log_c(about='pytest')
    logger.log_e(about='pytest')
    logger.log_i(about='pytest')
    logger.log_w(about='pytest')

    output = caplog.record_tuples

    assert len(output) == 6



def test_RobieLogger_cover(
    robie: 'Robie',
    caplog: LogCaptureFixture,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    :param caplog: pytest object for capturing log message.
    """

    logger = robie.logger
    childs = robie.childs
    clients = childs.clients


    thread = Thread(name='foo')

    client = clients['ircbot']


    logger.log_i(base=thread)
    logger.log_i(item=thread)
    logger.log_i(base=robie)
    logger.log_i(item=robie)
    logger.log_i(name=client)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from enconnect.discord import ClientEvent
from enconnect.discord.test import EVENTS

from pytest import raises

from ..client import DSCClient
from ..command import DSCCommand
from ..message import DSCMessage

if TYPE_CHECKING:
    from ....robie import Robie



def test_DSCMessage(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    model = DSCMessage

    client = clients['dscbot']

    assert isinstance(
        client, DSCClient)

    event = ClientEvent(
        client.client,
        EVENTS[0])


    item = model(
        client, event)


    attrs = lattrs(item)

    assert attrs == [
        'event',
        'client',
        'time']


    assert inrepr(
        'DSCMessage',
        item)

    with raises(TypeError):
        assert hash(item) > 0

    assert instr(
        'DSCMessage',
        item)


    assert item.time.since > 0

    assert item.client == client.name

    assert item.family == 'discord'

    assert item.kind == 'privmsg'

    assert item.event == event

    assert not item.isme


    assert event.type == (
        'MESSAGE_CREATE')
    assert event.opcode == 0
    assert event.data
    assert len(event.data) == 3
    assert event.seqno == 3
    assert len(event.original) == 4

    assert event.kind == 'privmsg'
    assert event.author == (
        'user', 'userid')
    assert event.recipient == (
        None, 'privid')
    assert event.message == (
        'Hello dscbot')



def test_DSCMessage_reply(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    model = DSCMessage

    client = clients['dscbot']

    assert isinstance(
        client, DSCClient)


    item = model(
        client,
        ClientEvent(
            client.client,
            EVENTS[0]))

    assert not item.isme


    reply = item.reply(
        robie, 'Hello')


    assert isinstance(
        reply, DSCCommand)


    assert reply.family == 'discord'

    assert reply.method == 'post'

    assert reply.path[-4:] == 'ages'

    assert not reply.params

    assert reply.json == {
        'content': 'Hello'}



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon.types import inrepr
from encommon.types import instr
from encommon.types import lattrs

from enconnect.irc import ClientEvent

from ..client import IRCClient
from ..command import IRCCommand
from ....robie.addons import RobieQueue

if TYPE_CHECKING:
    from ....robie import Robie
    from ....robie.models import RobieCommand  # noqa: F401
    from ....robie.models import RobieMessage  # noqa: F401



def test_IRCClient(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients


    client = clients['ircbot']

    assert isinstance(
        client,
        IRCClient)


    attrs = lattrs(client)

    assert attrs == [
        '_RobieChild__robie',
        '_RobieChild__name',
        '_RobieChild__params',
        '_IRCClient__client']


    assert inrepr(
        'client.IRCClient',
        client)

    assert hash(client) > 0

    assert instr(
        'client.IRCClient',
        client)


    client.validate()

    assert client.robie

    assert client.enable

    assert client.name == 'ircbot'

    assert client.family == 'irc'

    assert client.kind == 'client'

    assert client.params

    assert client.dumped



def test_IRCClient_message(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']

    assert isinstance(
        client, IRCClient)


    _queue = RobieQueue[
        'RobieMessage']

    queue: _queue = (
        RobieQueue(robie))


    _event = 'PING :123456789'

    event = ClientEvent(
        client.client, _event)

    client.put_message(
        queue, event)



def test_IRCClient_command(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']

    assert isinstance(
        client, IRCClient)


    _queue = RobieQueue[
        'RobieCommand']

    queue: _queue = (
        RobieQueue(robie))


    client.put_command(
        queue,
        'PRIVMSG # :Hello')



def test_IRCClient_compose(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    client = clients['ircbot']

    assert isinstance(
        client, IRCClient)


    citem = (
        client.compose(
            '#chan', 'message'))

    assert isinstance(
        citem, IRCCommand)


    assert citem.event == (
        'PRIVMSG #chan :message')



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from dataclasses import dataclass
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from enconnect.utils.http import _METHODS
from enconnect.utils.http import _PAYLOAD

from ...robie.models import RobieCommand

if TYPE_CHECKING:
    from ...robie.childs import RobieClient



@dataclass
class MTMCommand(RobieCommand):
    """
    Contain information for sharing using the Python queue.
    """

    method: _METHODS
    path: str
    params: Optional[_PAYLOAD]
    json: Optional[_PAYLOAD]


    def __init__(
        self,
        client: 'RobieClient',
        method: _METHODS,
        path: str,
        params: Optional[_PAYLOAD] = None,
        json: Optional[_PAYLOAD] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        # Translate to client.request

        self.method = method
        self.path = path
        self.params = params
        self.json = json

        super().__init__(client=client)


    @property
    def family(
        self,
    ) -> Literal['mattermost']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return 'mattermost'



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from threading import Event
from time import sleep as block_sleep
from typing import Any
from typing import TYPE_CHECKING

from encommon.times import Timer

from .members import RobieClients
from .members import RobiePlugins

if TYPE_CHECKING:
    from .robie import Robie
    from .params import RobieServiceParams



_CONGEST = list[tuple[str, int]]



class RobieService:
    """
    Multi-threaded service for processing the desired state.

    :param robie: Primary class instance for Chatting Robie.
    """

    __robie: 'Robie'

    __clients: RobieClients
    __plugins: RobiePlugins

    __timer: Timer
    __vacate: Event
    __cancel: Event
    __started: bool


    def __init__(
        self,
        robie: 'Robie',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        robie.logger.log_d(
            base=self,
            status='initial')

        self.__robie = robie

        self.__build_objects()

        self.__started = False

        robie.logger.log_i(
            base=self,
            status='created')


    def __build_objects(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        robie = self.__robie
        params = self.params

        respite = params.respite


        self.__clients = (
            RobieClients(robie))

        self.__plugins = (
            RobiePlugins(robie))


        self.__timer = Timer(
            respite.health,
            start='min')

        self.__vacate = Event()
        self.__cancel = Event()


    @property
    def robie(
        self,
    ) -> 'Robie':
        """
        Return the Robie instance to which this instance belongs.

        :returns: Robie instance to which this instance belongs.
        """

        return self.__robie


    @property
    def params(
        self,
    ) -> 'RobieServiceParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        robie = self.__robie
        params = robie.params

        return params.service


    @property
    def clients(
        self,
    ) -> RobieClients:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__clients


    @property
    def plugins(
        self,
    ) -> RobiePlugins:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__plugins


    @property
    def running(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined running.

        :returns: List of threads which are determined running.
        """

        running: list[str] = []

        clients = self.__clients
        plugins = self.__plugins

        running.extend(
            clients.running)

        running.extend(
            plugins.running)

        return sorted(running)


    @property
    def zombies(
        self,
    ) -> list[str]:
        """
        Return the list of threads which are determined zombies.

        :returns: List of threads which are determined zombies.
        """

        zombies: list[str] = []

        clients = self.__clients
        plugins = self.__plugins

        zombies.extend(
            clients.zombies)

        zombies.extend(
            plugins.zombies)

        return sorted(zombies)


    @property
    def congest(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        congest: _CONGEST = []

        clients = self.__clients
        plugins = self.__plugins

        congest.extend(
            clients.congest)

        congest.extend(
            plugins.congest)

        return sorted(congest)


    @property
    def enqueue(
        self,
    ) -> _CONGEST:
        """
        Return the list of congested threads and members queues.

        :returns: List of congested threads and members queues.
        """

        enqueue: _CONGEST = []

        clients = self.__clients
        plugins = self.__plugins

        enqueue.extend(
            clients.enqueue)

        enqueue.extend(
            plugins.enqueue)

        return sorted(enqueue)


    def start(
        self,
    ) -> None:
        """
        Start the various threads within the Homie class object.
        """

        robie = self.__robie
        started = self.__started

        if started is True:
            return None

        self.__started = True

        robie.logger.log_i(
            base=self,
            status='starting')

        self.__plugins.start()
        self.__clients.start()

        robie.logger.log_i(
            base=self,
            status='started')


    def operate(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        robie = self.__robie

        clients = self.__clients
        plugins = self.__plugins

        while self.running:

            self.operate_clients()
            self.operate_plugins()

            clients.operate()
            plugins.operate()

            block_sleep(0.05)

            self.operate_healths()

        robie.logger.log_i(
            base=self,
            status='vacated')


    def operate_clients(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        robie = self.__robie

        vacate = self.__vacate
        clients = self.__clients
        plugins = self.__plugins

        source = plugins.cqueue
        target = clients.cqueue


        while not source.empty:

            citem = source.get()

            if vacate.is_set():
                continue

            target.put(citem)

            robie.printer(citem)


    def operate_plugins(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        robie = self.__robie

        vacate = self.__vacate
        clients = self.__clients
        plugins = self.__plugins

        source = clients.mqueue
        target = plugins.mqueue


        while not source.empty:

            mitem = source.get()

            if vacate.is_set():
                continue

            target.put(mitem)

            robie.printer(mitem)


    def operate_healths(
        self,
    ) -> None:
        """
        Perform the operation related to Homie service members.
        """

        timer = self.__timer

        if timer.pause():
            return None

        vacate = self.__vacate
        cancel = self.__cancel

        self.check_congest()

        if vacate.is_set():
            return None

        if cancel.is_set():
            return None

        if self.check_zombies():
            self.stop()


    def check_zombies(
        self,
    ) -> bool:
        """
        Return the boolean indicating while threads are zombies.

        :returns: Boolean indicating while threads are zombies.
        """

        robie = self.__robie
        zombies = self.zombies

        for name in zombies:

            robie.logger.log_c(
                thread=name,
                status='zombie')

        return len(zombies) != 0


    def check_congest(
        self,
    ) -> bool:
        """
        Return the boolean indicating when queues are congested.

        :returns: Boolean indicating when queues are congested.
        """

        robie = self.__robie
        congest = self.congest

        for name, count in congest:

            robie.logger.log_e(
                queue=name,
                status='congest',
                count=count)

        return len(congest) != 0


    def soft(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Stop the various threads within the Homie class object.

        :param kwargs: Keyword arguments ignored by the method.
        :param args: Positional arguments ignored by the method.
        """

        robie = self.__robie
        started = self.__started
        vacate = self.__vacate

        if started is False:
            return None

        if vacate.is_set():
            return self.stop()

        robie.logger.log_i(
            base=self,
            status='vacating')

        vacate.set()

        self.__clients.stop()
        self.__plugins.soft()


    def stop(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Stop the various threads within the Homie class object.

        :param kwargs: Keyword arguments ignored by the method.
        :param args: Positional arguments ignored by the method.
        """

        robie = self.__robie
        started = self.__started
        vacate = self.__vacate
        cancel = self.__cancel

        if started is False:
            return None

        if cancel.is_set():
            return None

        vacate.set()
        cancel.set()

        robie.logger.log_i(
            base=self,
            status='stopping')

        self.__clients.stop()
        self.__plugins.stop()

        robie.logger.log_i(
            base=self,
            status='stopped')



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING

from .params import StatusPluginParams
from ...robie.childs import RobiePlugin

if TYPE_CHECKING:
    from ...robie.threads import RobieThread



class StatusPlugin(RobiePlugin):
    """
    Integrate with the Robie routine and perform operations.

    .. note::
       This plugin responds to inquiries about Robie status.

    :param robie: Primary class instance for Chatting Robie.
    """


    def validate(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """


    def operate(
        self,
        thread: 'RobieThread',
    ) -> None:
        """
        Perform the operation related to Homie service threads.

        :param thread: Child class instance for Chatting Robie.
        """

        robie = thread.robie
        mqueue = thread.mqueue
        member = thread.member
        cqueue = member.cqueue
        params = self.params

        assert isinstance(
            params, StatusPluginParams)

        command = params.command
        match: Optional[str]

        kinds = ['privmsg', 'chanmsg']


        while not mqueue.empty:

            mitem = mqueue.get()

            time = mitem.time
            kind = mitem.kind
            isme = mitem.isme
            family = mitem.family

            if kind not in kinds:
                continue

            if isme is True:
                continue

            if time.since > 15:
                continue  # NOCVR

            event = getattr(
                mitem, 'event')


            match = None

            if family == 'discord':
                match = command.dsc

            if family == 'irc':
                match = command.irc

            if family == 'mattermost':
                match = command.mtm

            if match is None:
                continue  # NOCVR


            message = event.message

            if message != match:
                continue  # NOCVR


            citem = mitem.reply(
                robie,
                'All good here')

            cqueue.put(citem)



"""
Functions and routines associated with Enasis Network Chatting Robie.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...robie import Robie



def test_RobieChild_cover(
    robie: 'Robie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param robie: Primary class instance for Chatting Robie.
    """

    childs = robie.childs
    clients = childs.clients

    child1 = clients['ircbot']
    child2 = clients['dscbot']


    sort = [child1, child2]

    assert sorted(sort) == [
        child2, child1]
